import typing

Flake8Error = typing.NamedTuple(
    'Flake8Error', [
        ('line_number', int),
        ('offset', int),
        ('text', str),
        ('checker_cls', type),
    ]
)

AAAError = typing.NamedTuple('AAAError', [
    ('line_number', int),
    ('offset', int),
    ('text', str),
])


class Flake8AAAException(Exception):
    pass


class TokensNotLoaded(Flake8AAAException):
    """
    `Checker.all_funcs()` was called before `ast_tokens` was populated. Usually
    this is done by `Checker.load()`.
    """


class EmptyBlock(Flake8AAAException):
    """
    Block has no nodes.
    """


class ValidationError(Flake8AAAException):
    """
    Attributes:
        line_number (int)
        offset (int)
        text (str)
    """

    def __init__(self, line_number, offset, text):
        self.line_number = line_number
        self.offset = offset
        self.text = text

    def to_flake8(self, checker_cls: type) -> Flake8Error:
        """
        Args:
            checker_cls: Class performing the check to be passed back to
                flake8.
        """
        return Flake8Error(
            line_number=self.line_number,
            offset=self.offset,
            text=self.text,
            checker_cls=checker_cls,
        )

    def to_aaa(self) -> AAAError:
        return AAAError(
            line_number=self.line_number,
            offset=self.offset,
            text=self.text,
        )



import ast
from typing import Iterable, List, Tuple, Type, TypeVar

from .exceptions import EmptyBlock
from .helpers import filter_arrange_nodes, get_first_token, get_last_token
from .types import LineType

_Block = TypeVar('_Block', bound='Block')


class Block:
    """
    An Arrange, Act or Assert block of code as parsed from the test function.

    Note:
        This may just become the Act Block *AND* since the Act Block is just a
        single node, this might not even be required.

    Args:
        nodes: Nodes that make up this block.
        line_type: Type of line that this blocks writes into the line markers
            instance for the function.

    Notes:
        * Blocks with no nodes are allowed (at the moment).
    """

    def __init__(self, nodes: Iterable[ast.AST], lt: LineType) -> None:
        self.nodes = tuple(nodes)
        self.line_type = lt

    @classmethod
    def build_act(cls: Type[_Block], node: ast.stmt) -> _Block:
        """
        Act block is a single node.
        """
        return cls([node], LineType.act)

    @classmethod
    def build_arrange(cls: Type[_Block], nodes: List[ast.stmt], act_block_first_line: int) -> _Block:
        """
        Arrange block is all non-pass and non-docstring nodes before the Act
        block start.

        Args:
            nodes: Body of test function / method.
            act_block_first_line
        """
        return cls(filter_arrange_nodes(nodes, act_block_first_line), LineType.arrange)

    def get_span(self, first_line_no: int) -> Tuple[int, int]:
        """
        Args:
            first_line_no: First line number of Block. Used to calculate
                relative line numbers.

        Returns:
            First and last line covered by this block, counted relative to the
            start of the Function.

        Raises:
            EmptyBlock: when block has no nodes
        """
        if not self.nodes:
            raise EmptyBlock(f'span requested from {self.line_type} block with no nodes')
        # start and end are (<line number>, <indent>) pairs, so just the line
        # numbers are picked out.
        return (
            get_first_token(self.nodes[0]).start[0] - first_line_no,
            get_last_token(self.nodes[-1]).end[0] - first_line_no,
        )



import ast
from typing import List, Type, TypeVar

from .helpers import get_first_token, node_is_pytest_raises, node_is_result_assignment, node_is_unittest_raises
from .types import ActNodeType

AN = TypeVar('AN', bound='ActNode')  # Place holder for ActNode instances


class ActNode:
    """
    Attributes:
        node
        block_type
    """

    def __init__(self, node: ast.stmt, block_type: ActNodeType) -> None:
        """
        Args:
            node
            block_type
        """
        self.node = node
        self.block_type = block_type

    @classmethod
    def build_body(cls: Type[AN], body: List[ast.stmt]) -> List:
        """
        Note:
            Return type is probably ``-> List[AN]``, but can't get it to pass.
        """
        act_nodes: List[ActNode] = []
        for child_node in body:
            act_nodes += ActNode.build(child_node)
        return act_nodes

    @classmethod
    def build(cls: Type[AN], node: ast.stmt) -> List[AN]:
        """
        Starting at this ``node``, check if it's an act node. If it's a context
        manager, recurse into child nodes.

        Returns:
            List of all act nodes found.
        """
        if node_is_result_assignment(node):
            return [cls(node, ActNodeType.result_assignment)]
        if node_is_pytest_raises(node):
            return [cls(node, ActNodeType.pytest_raises)]
        if node_is_unittest_raises(node):
            return [cls(node, ActNodeType.unittest_raises)]

        # Check if line marked with '# act'
        if get_first_token(node).line.strip().endswith('# act'):
            return [cls(node, ActNodeType.marked_act)]

        # Recurse (downwards) if it's a context manager
        if isinstance(node, ast.With):
            return cls.build_body(node.body)

        return []



from typing import Generator, List, Optional

from .exceptions import AAAError, ValidationError
from .helpers import first_non_blank_char
from .types import LineType


class LineMarkers:
    """
    Marks each line of a test function with the ``LineType`` assigned to that
    line.

    Attributes:
        lines: Lines of code that make up test. Used to calculate offset in
            raised error.
        types: List of types for each line in test.
        fn_offset: First line number of test so that line type indices can be
            converted back to line numbers in exceptions.
    """

    def __init__(self, lines: List[str], fn_offset: int) -> None:
        self.types = [LineType.unprocessed] * len(lines)
        self.lines = lines
        self.fn_offset = fn_offset

    def set(self, index: int, value: LineType) -> bool:
        """
        Extended version of setitem to assert that rules for setting a line are
        met:
            * Existing line is unprocessed - line is set as new type.
            * Existing line is blank line or comment and new line is function
                defintion or an AAA block - line setting is ignored because
                comments and blank lines can appear in func defs and blocks.

        Returns:
            An unprocessed line was replaced with a new line type.

        Raises:
            ValidationError: AAA99 marking caused a collision.
            ValueError: passed ``value`` is unprocessed line or is not a
                LineType.
        """
        if not isinstance(value, LineType):
            raise ValueError(f'"{value}" for line index {index} is not LineType')
        if value is LineType.unprocessed:
            raise ValueError(f'Can not revert line index {index} to "{value}"')
        current_type = self.types[index]
        if current_type is LineType.blank_line or current_type is LineType.comment:
            return False
        if current_type is not LineType.unprocessed:
            line_num = index + self.fn_offset
            raise ValidationError(
                line_num,
                1,
                f'AAA99 collision when marking line {line_num} (index={index}) as {value}, was already {current_type}',
            )
        self.types[index] = value
        return True

    def update(self, a: int, b: int, line_type: LineType) -> int:
        """
        Updates line types from index ``a`` to index ``b`` inclusively. Indexes
        are relative.

        Args:
            a: First line index.
            b: Last line index.
            line_type: New type of line.

        Returns:
            Number of lines updated. This may not be equal to ``b - a + 1``
                because lines that are blank are skipped.

        Raises:
            ValidationError: A special error on collision. This prevents Flake8
                from crashing because it is converted to a Flake8 error tuple,
                but it indicates to the user that something went wrong with
                processing the function.
        """
        count = 0

        for i in range(a, b + 1):
            if self.set(i, line_type):
                count += 1

        return count

    def previous(self, num: int) -> Optional[LineType]:
        """
        Returns:
            Previous line's type, relative to `num`. Returns `None` if we're at
            the first line.
        """
        try:
            return self.types[num - 1]
        except IndexError:
            return None

    def next(self, num) -> Optional[LineType]:
        """
        Returns:
            Next line's type, relative to `num`. Returns `None` if we're at the
            last line.
        """
        try:
            return self.types[num + 1]
        except IndexError:
            return None

    def check_arrange_act_spacing(self) -> Generator[AAAError, None, None]:
        """
        * When no spaces found, point error at line above act block
        * When too many spaces found, point error at 2nd blank line
        """
        yield from self.check_block_spacing(
            LineType.arrange,
            LineType.act,
            'AAA03 expected 1 blank line before Act block, found {}',
        )

    def check_act_assert_spacing(self) -> Generator[AAAError, None, None]:
        """
        * When no spaces found, point error at line above assert block
        * When too many spaces found, point error at 2nd blank line
        """
        yield from self.check_block_spacing(
            LineType.act,
            LineType._assert,
            'AAA04 expected 1 blank line before Assert block, found {}',
        )

    def check_block_spacing(
        self,
        first_block_type: LineType,
        second_block_type: LineType,
        error_message: str,
    ) -> Generator[AAAError, None, None]:
        """
        Checks there is a clear single line between ``first_block_type`` and
        ``second_block_type``.

        Note:
            Is tested via ``check_arrange_act_spacing()`` and
            ``check_act_assert_spacing()``.
        """
        numbered_lines = list(enumerate(self.types))
        first_block_lines = filter(lambda l: l[1] is first_block_type, numbered_lines)
        try:
            first_block_lineno = list(first_block_lines)[-1][0]
        except IndexError:
            # First block has no lines
            return

        second_block_lines = filter(lambda l: l[1] is second_block_type, numbered_lines)
        try:
            second_block_lineno = next(second_block_lines)[0]
        except StopIteration:
            # Second block has no lines
            return

        blank_lines = [
            bl for bl in numbered_lines[first_block_lineno + 1:second_block_lineno] if bl[1] is LineType.blank_line
        ]

        if not blank_lines or len(blank_lines) != 1:
            # Point at first line of second block
            yield self.build_error(
                line_index=second_block_lineno,
                text=error_message.format('none' if not blank_lines else len(blank_lines)),
            )
            return

    def check_blank_lines(self) -> Generator[AAAError, None, None]:
        checked_blocks = (LineType.func_def, LineType.arrange, LineType.act, LineType._assert)
        for num, line_type in enumerate(self.types):
            if (
                line_type is LineType.blank_line and self.types[num - 1] in checked_blocks
                and self.types[num - 1] == self.types[num + 1]
            ):
                yield self.build_error(
                    line_index=num,
                    text='AAA05 blank line in block',
                )

    def check_comment_in_act(self) -> Generator[AAAError, None, None]:
        for num, line_type in enumerate(self.types):
            if line_type is LineType.comment:
                if self.previous(num) == LineType.act or self.next(num) == LineType.act:
                    yield self.build_error(
                        line_index=num,
                        text='AAA06 comment in Act block',
                    )

    def build_error(self, line_index: int, text: str) -> AAAError:
        """
        Calculate the offset of the error based on the first non-blank
        character of the line.
        """
        return AAAError(
            line_number=self.fn_offset + line_index,
            offset=first_non_blank_char(self.lines[line_index]),
            text=text,
        )



from enum import Enum, unique

ActNodeType = Enum(
    'ActNodeType',
    (
        'marked_act '  # Marked with "# act"
        'pytest_raises '  # Wrapped in "pytest.raises" context manager.
        'result_assignment '  # Simple "result = "
        'unittest_raises '  # Wrapped in unittest's "assertRaises" context manager.
    ),
)


@unique
class LineType(Enum):
    # Act
    act = 'ACT'
    # Arrange
    arrange = 'ARR'
    # Assert
    _assert = 'ASS'
    # Blank line
    blank_line = 'BL '
    # Function definition
    func_def = 'DEF'
    # Comments
    comment = 'CMT'
    # Not processed line
    unprocessed = 'QQQ'

    def __str__(self) -> str:
        return '???' if self == self.unprocessed else self.value



import ast
import tokenize
from typing import Generator, List, Optional

from asttokens.util import Token as ASTToken

from .act_node import ActNode
from .block import Block
from .exceptions import AAAError, EmptyBlock, ValidationError
from .helpers import format_errors, function_is_noop, get_first_token, get_last_token, line_is_comment
from .line_markers import LineMarkers
from .types import ActNodeType, LineType


class Function:
    """
    Attributes:
        act_block: Block wrapper around the single Act node.
        act_node: Act Node for the test. This is the node of the AST that looks
            like the Action. Distinguish between this and the Act Block - the
            Act Block can be larger than just the node, mainly because it could
            be wrapped in context managers. Defaults to ``None``.
        arrange_block: Arrange block for this test. Defaults to ``None``.
        assert_block: Assert block. Defaults to ``None``.
        _errors: List of errors for this Function. Defaults to ``None`` when
            Function has not been checked. Empty list ``[]`` means that the
            Function has been checked and is free of errors.
        first_line_no: Line number of the first token in the test. Used to hop
            to and from relative line numberings.
        lines: Slice of the file lines that make up this test function /
            method.
        line_markers: Line-wise marking for this function.
        node: AST for the test function / method.
        tokens: Slice of the file's tokens that make up this test function.

    Note:
        "line number" means the number of the line in the file (the usual
        definition). "offset" means the number of the line in the test relative
        to the test definition.
    """

    def __init__(self, node: ast.FunctionDef, file_lines: List[str], file_tokens: List[ASTToken]):
        """
        Args:
            node
            file_lines: Lines of file under test.
            file_tokens: Tokens for file passed by Flake8.
        """
        self.node = node
        self.first_line_no: int = get_first_token(self.node).start[0]
        end: int = get_last_token(self.node).end[0]
        self.lines: List[str] = file_lines[self.first_line_no - 1:end]
        self.arrange_block: Optional[Block] = None
        self.act_node: Optional[ActNode] = None
        self.act_block: Optional[Block] = None
        self.assert_block: Optional[Block] = None
        self.line_markers = LineMarkers(self.lines, self.first_line_no)
        self.tokens = file_tokens

    def __str__(self, errors: Optional[List[AAAError]] = None) -> str:
        out = '------+------------------------------------------------------------------------\n'
        for i, line in enumerate(self.lines):
            out += '{line_no:>2} {block}|{line}'.format(
                line_no=i + self.first_line_no,
                block=str(self.line_markers.types[i]),
                line=line,
            )
            if errors:
                for error in errors:
                    if error[0] == i + self.first_line_no:
                        out += '       {}^ {}\n'.format(error[1] * ' ', error[2])
        out += '------+------------------------------------------------------------------------\n'
        if errors is not None:
            out += format_errors(len(errors))
        return out

    def check_all(self) -> Generator[AAAError, None, None]:
        """
        Run everything required for checking this test.

        Returns:
            A generator of errors.

        Raises:
            ValidationError: A non-recoverable linting error is found.
        """
        # Function def
        if function_is_noop(self.node):
            return

        self.mark_bl()
        self.mark_comments()
        self.mark_def()

        self.mark_act()
        self.mark_arrange()
        self.mark_assert()

        yield from self.line_markers.check_arrange_act_spacing()
        yield from self.line_markers.check_act_assert_spacing()
        yield from self.line_markers.check_blank_lines()
        yield from self.line_markers.check_comment_in_act()

    def mark_bl(self) -> int:
        """
        Mark unprocessed lines that have no content and no string nodes
        covering them as blank line BL.

        Returns:
            Number of blank lines found.
        """
        counter = 0
        previous = None
        for t in self.tokens:
            if t.type == tokenize.NL:
                assert previous is not None, "Unexpected NL token before any other tokens seen"
                if previous.type == tokenize.NL or previous.type == tokenize.NEWLINE:
                    self.line_markers.types[t.start[0] - self.first_line_no] = LineType.blank_line
                    counter += 1
            previous = t

        return counter

    def mark_comments(self) -> int:
        """
        Mark unprocessed lines that are just `#` comments as CMT.

        Returns:
            Number of comment lines found.
        """
        counter = 0
        previous = None
        for t in self.tokens:
            if t.type == tokenize.COMMENT:
                assert previous is not None, "Unexpected COMMENT token before any other tokens seen"
                if previous.type == tokenize.NL or previous.type == tokenize.NEWLINE:
                    self.line_markers.types[t.start[0] - self.first_line_no] = LineType.comment
                    counter += 1
            previous = t

        return counter

    def mark_def(self) -> int:
        """
        Marks up this Function's definition lines (including decorators) into
        the ``line_markers`` attribute.

        Returns:
            Number of lines found for the definition.

        Note:
            Does not spot the closing ``):`` of a function when it occurs on
            its own line.

        Note:
            Can not use ``helpers.build_footprint()`` because function nodes
            cover the whole function. In this case, just the def lines are
            wanted with any decorators.
        """
        first_index = get_first_token(self.node).start[0] - self.first_line_no  # Should usually be 0
        try:
            end_token = get_last_token(self.node.args.args[-1])
        except IndexError:
            # Fn has no args, so end of function is the fn def itself...
            end_token = get_first_token(self.node)
        last_index = end_token.end[0] - self.first_line_no
        self.line_markers.update(first_index, last_index, LineType.func_def)
        return last_index - first_index + 1

    def mark_act(self) -> int:
        """
        Finds Act node, calculates its span and marks the associated lines in
        ``line_markers``.

        Returns:
            Number of lines covered by the Act block (used for debugging /
            testing only). Includes any comment or blank lines already marked
            inside the block's span.

        Raises:
            ValidationError: Muliple possible fatal errors:
                * AAA01 no act block found.
                * AAA02 multiple act blocks found.
                * AAA99 marking caused a collision.
        """
        # Load act block and kick out when none is found
        self.act_node = self.load_act_node()
        self.act_block = Block.build_act(self.act_node.node)
        # Get relative line numbers of Act block footprint
        # TODO store first and last line numbers in Block - use them instead of
        # asking for span.
        first_index, last_index = self.act_block.get_span(self.first_line_no)
        self.line_markers.update(first_index, last_index, LineType.act)
        return last_index - first_index + 1

    def mark_arrange(self) -> int:
        """
        Mark all lines of code *before* the Act block as Arrange in
        ``line_markers``. Location of Act block is sniffed from
        ``line_markers``.

        Returns:
            Number of lines covered by the Arrange block (used for debugging /
            testing only).

        Raises:
            ValidationError: Marking caused a collision.
            ValueError: No Act block has been marked.
        """
        # TODO get this from self.act_block
        act_block_first_index = self.line_markers.types.index(LineType.act)
        act_block_first_line_number = act_block_first_index + self.first_line_no
        arrange_block = Block.build_arrange(self.node.body, act_block_first_line_number)

        # First and lass offsets of Arrange block - if block is empty, then
        # work is done.
        try:
            first_index, last_index = arrange_block.get_span(self.first_line_no)
        except EmptyBlock:
            return 0

        # Prevent overhanging arrangement, for example in context manager. Stop
        # at line before Act block first line offset.
        return self.line_markers.update(
            first_index,
            min(last_index, act_block_first_index - 1),
            LineType.arrange,
        )

    def mark_assert(self) -> int:
        """
        Mark all lines of code *after* the Act block as Assert in
        ``line_markers``.

        Returns:
            Number of lines covered by the Assert block (used for debugging /
            testing only).

        Raises:
            ValidationError: AAA99 marking caused a collision.
        """
        count = 0
        # TODO get this from self.act_block
        # TODO keep length of function around instead of counting lines
        act_block_last_index = len(self.line_markers.types) - 1 - self.line_markers.types[::-1].index(LineType.act)

        # Starting from the line after the last line of Act block, to the end
        # of the test, mark everything that's unprocessed, and not a comment,
        # as an Assert block item.
        # TODO keep length of function around instead of counting lines
        for offset in range(act_block_last_index + 1, len(self.line_markers.types)):
            if self.line_markers.types[offset] == LineType.unprocessed and not line_is_comment(self.lines[offset]):
                count += 1
                self.line_markers.types[offset] = LineType._assert

        return count

    def load_act_node(self) -> ActNode:
        """
        Raises:
            ValidationError:
                * AAA01 no act block found.
                * AAA02 multiple act blocks found.
        """
        act_nodes = ActNode.build_body(self.node.body)

        if not act_nodes:
            raise ValidationError(self.first_line_no, self.node.col_offset, 'AAA01 no Act block found in test')

        # Allow `pytest.raises` and `self.assertRaises()` in assert nodes - if
        # any of the additional nodes are `pytest.raises`, then raise
        for a_n in act_nodes[1:]:
            if a_n.block_type in [ActNodeType.marked_act, ActNodeType.result_assignment]:
                raise ValidationError(
                    self.first_line_no,
                    self.node.col_offset,
                    'AAA02 multiple Act blocks found in test',
                )

        return act_nodes[0]



import ast
import typing

from .checker import Checker
from .exceptions import ValidationError


def do_command_line(infile: typing.IO[str]) -> int:
    """
    Currently a small stub to create an instance of Checker for the passed
    ``infile`` and run its test functions through linting.

    Args:
        infile

    Returns:
        int: Number of flake8 errors raised.
    """
    lines = infile.readlines()
    tree = ast.parse(''.join(lines))
    checker = Checker(tree, lines, infile.name)
    checker.load()
    num_errors = 0

    for func in checker.all_funcs(skip_noqa=True):
        try:
            errors = list(func.check_all())
        except ValidationError as error:
            errors = [error.to_aaa()]
        print(func.__str__(errors), end='')
        num_errors += len(errors)

    print('======+========================================================================')
    if num_errors:
        print('        FAILED with {} ERROR{}'.format(num_errors, 'S' if num_errors > 1 else ''))
    else:
        print('        PASSED!')

    return num_errors



from ast import AST
from typing import Generator, List, Optional, Tuple

import asttokens

from .__about__ import __short_name__, __version__
from .exceptions import TokensNotLoaded, ValidationError
from .function import Function
from .helpers import find_test_functions, is_test_file


class Checker:
    """
    Attributes:
        ast_tokens: Tokens for the file.
        filename: Name of file under check.
        lines
        tree: Tree passed from flake8.
    """

    name = __short_name__
    version = __version__

    def __init__(self, tree: AST, lines: List[str], filename: str):
        self.tree = tree
        self.lines = lines
        self.filename = filename
        self.ast_tokens: Optional[asttokens.ASTTokens] = None

    def load(self) -> None:
        self.ast_tokens = asttokens.ASTTokens(''.join(self.lines), tree=self.tree)

    def all_funcs(self, skip_noqa: bool = False) -> Generator[Function, None, None]:
        """
        Note:
            Checker is responsible for slicing the tokens passed to the
            Function, BUT the function is reponsible for slicing the lines.
            This is a bit strange - instead the lines should be sliced here and
            passed in so that the Function only receives data about itself.

        Raises:
            TokensNotLoaded: On fetching first value, when `load()` has not
                been called to populate `ast_tokens`.
        """
        if self.ast_tokens is None:
            raise TokensNotLoaded("ast_tokens is `None`")
        for f in find_test_functions(self.tree, skip_noqa=skip_noqa):
            yield Function(f, self.lines, list(self.ast_tokens.get_tokens(f, include_extra=True)))

    def run(self) -> Generator[Tuple[int, int, str, type], None, None]:
        """
        Yields:
            tuple (line_number, offset, text, check)
        """
        if is_test_file(self.filename):
            self.load()
            for func in self.all_funcs():
                try:
                    for error in func.check_all():
                        yield (error.line_number, error.offset, error.text, Checker)
                except ValidationError as error:
                    yield error.to_flake8(Checker)



__short_name__ = 'aaa'
__iam__ = 'flake8-{}'.format(__short_name__)
__version__ = '0.11.0'

__author__ = 'James Cooke'
__copyright__ = '2018 - 2020, {}'.format(__author__)

__description__ = 'A Flake8 plugin that checks Python tests follow the Arrange-Act-Assert pattern'
__email__ = 'github@jamescooke.info'



import ast
import io
import os
import re
import tokenize
from typing import List

from asttokens.util import Token

test_file_pattern = re.compile(r'test(_.*|s)?\.py$')


def is_test_file(filename: str) -> bool:
    """
    Check that path to file being checked passed by flake8 looks like a pytest
    test file.
    """
    return bool(test_file_pattern.match(os.path.basename(filename)))


def first_non_blank_char(line: str) -> int:
    for pos, char in enumerate(line):
        if not char.isspace():
            return pos
    return 0


class TestFuncLister(ast.NodeVisitor):
    """
    Helper to walk the ast Tree and find functions that looks like tests.
    Matching function nodes are kept in ``_found_func`` attr.
    """

    def __init__(self, skip_noqa: bool):
        super(TestFuncLister, self).__init__()
        self.skip_noqa = skip_noqa
        self._found_funcs: List[ast.FunctionDef] = []

    def visit_FunctionDef(self, node):
        if node.name.startswith('test'):
            if not self.skip_noqa or not node.first_token.line.strip().endswith('# noqa'):
                self._found_funcs.append(node)

    def get_found_funcs(self) -> List[ast.FunctionDef]:
        return self._found_funcs


def find_test_functions(tree: ast.AST, skip_noqa: bool = False) -> List[ast.FunctionDef]:
    """
    Collect functions that look like tests.

    Args:
        tree
        skip_noqa: Flag used by command line debugger to skip functions that
            are marked with "# noqa". Defaults to ``False``.
    """
    function_finder = TestFuncLister(skip_noqa)
    function_finder.visit(tree)
    return function_finder.get_found_funcs()


def node_is_result_assignment(node: ast.AST) -> bool:
    """
    Args:
        node: An ``ast`` node.

    Returns:
        bool: ``node`` corresponds to the code ``result =``, assignment to the
        ``result `` variable.
    """
    if isinstance(node, ast.Assign):
        return len(node.targets) == 1 and isinstance(node.targets[0], ast.Name) and node.targets[0].id == "result"

    if isinstance(node, ast.AnnAssign):
        return node.target.id == "result"  # type: ignore

    return False


def node_is_pytest_raises(node: ast.AST) -> bool:
    """
    Args:
        node: An ``ast`` node, augmented with ASTTokens

    Returns:
        bool: ``node`` corresponds to a With node where the context manager is
        ``pytest.raises``.
    """
    return isinstance(node, ast.With) and get_first_token(node).line.strip().startswith('with pytest.raises')


def node_is_unittest_raises(node: ast.AST) -> bool:
    """
    ``node`` corresponds to a With node where the context manager is unittest's
    ``self.assertRaises``.
    """
    return isinstance(node, ast.With) and get_first_token(node).line.strip().startswith('with self.assertRaises')


def node_is_noop(node: ast.AST) -> bool:
    """
    Node does nothing.
    """
    return isinstance(node.value, ast.Str) if isinstance(node, ast.Expr) else isinstance(node, ast.Pass)


def function_is_noop(function_node: ast.FunctionDef) -> bool:
    """
    Function does nothing - is just ``pass`` or docstring.
    """
    return all(node_is_noop(n) for n in function_node.body)


def format_errors(num_errors: int) -> str:
    """
    Formats a Function's errors for command line use.
    """
    if num_errors == 1:
        return '    1 | ERROR\n'
    return ' {:>4} | ERRORS\n'.format(num_errors)


def get_first_token(node: ast.AST) -> Token:
    """
    Wrapper to solve typing errors. mypy complains that ``ast.AST`` has no
    property ``first_token`` or ``last_token``. That's because these are added
    by the asttokens library. For now, this ignoring of type, which I think is
    required to get mypy to pass at this time, is encapsulated in this helper
    function.
    """
    return node.first_token  # type: ignore


def get_last_token(node: ast.AST) -> Token:
    """
    Performs same purpose as get_first_token.
    """
    return node.last_token  # type: ignore


def filter_arrange_nodes(nodes: List[ast.stmt], act_block_first_line_number: int) -> List[ast.stmt]:
    """
    Args:
        act_block_first_line_number: Real line number of Act block first line.

    Returns:
        All nodes are not docstrings or ``pass`` that start before the Act
        block's first line.
    """
    return [
        node for node in nodes if node.lineno < act_block_first_line_number and not isinstance(node, ast.Pass)
        and not (isinstance(node, ast.Expr) and isinstance(node.value, ast.Str))
    ]


def line_is_comment(line: str) -> bool:
    """
    Helper for checking that a single line is a comment. Will be replaced by a
    complete `find_comment_lines()` helper in #148. Could also use `tokens`
    from Flake8.
    """
    # TODO use existing tokens
    try:
        first_token = next(tokenize.generate_tokens(io.StringIO(line).readline))
    except tokenize.TokenError:
        # Assume that a token error happens because this is *not* a comment
        return False
    return first_token.type == tokenize.COMMENT



# Dirty hack to allow for direct running of flake8-aaa from the command line,
# bypassing flake8's harness. Goal of this is to allow for debugging of
# analysis, especially around how flake8-aaa has assigned lines of code to
# particular blocks.

import argparse
import sys

from .command_line import do_command_line


def main() -> int:
    parser = argparse.ArgumentParser(description='flake8-aaa command line debug')
    parser.add_argument('infile', type=argparse.FileType('r'), help='File to be linted')
    args = parser.parse_args()
    result = do_command_line(args.infile)
    args.infile.close()
    return result


if __name__ == '__main__':
    sys.exit(main())



import os

from setuptools import find_packages, setup

basedir = os.path.dirname(__file__)


def readme():
    with open(os.path.join(basedir, 'README.rst')) as f:
        return f.read()


about = {}
with open(os.path.join(basedir, 'src', 'flake8_aaa', '__about__.py')) as f:
    exec(f.read(), about)  # yapf: disable

setup(
    # --- META ---
    name=about['__iam__'],
    version=about['__version__'],
    description=about['__description__'],
    license='MIT',
    long_description=readme(),
    author='James Cooke',
    author_email='github@jamescooke.info',
    url='https://github.com/jamescooke/flake8-aaa',

    # --- Python ---
    packages=find_packages('src'),
    package_dir={'': 'src'},
    python_requires='>=3.6, <4',
    install_requires=[
        'asttokens >= 2',
    ],
    entry_points={
        'flake8.extension': [
            'AAA = flake8_aaa:Checker',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Flake8',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python',
    ],
    zip_safe=False,
)



import ast
from abc import ABCMeta, abstractmethod
from typing import List  # noqa: F401

from .types import LineType


class MultiNodeBlock(metaclass=ABCMeta):

    @property
    @abstractmethod
    def line_type(self) -> LineType:
        pass

    def __init__(self) -> None:
        self.nodes: List[ast.AST] = []

    @abstractmethod
    def add_node(self, node: ast.AST) -> bool:
        pass
